REPOSITORY_DEPENDENCY_DEFINITION_FILENAME = "repository_dependencies.xml"
REPOSITORY_SUITE_DEFINITION = "repository_suite_definition"
TOOL_DEPENDENCY_DEFINITION = "tool_dependency_definition"
TOOL_DEPENDENCY_DEFINITION_FILENAME = "tool_dependencies.xml"
UNRESTRICTED = "unrestricted"

types = [UNRESTRICTED, TOOL_DEPENDENCY_DEFINITION, REPOSITORY_SUITE_DEFINITION]

__all__ = (
    "REPOSITORY_DEPENDENCY_DEFINITION_FILENAME",
    "REPOSITORY_SUITE_DEFINITION",
    "TOOL_DEPENDENCY_DEFINITION",
    "TOOL_DEPENDENCY_DEFINITION_FILENAME",
    "UNRESTRICTED",
    "types",
)
